@define-color colorPrimary mix(#7239b3, @titlebar_color, 0.2);
@define-color colorAccent #8c53cb;
@define-color selected_bg_color @colorAccent;
@define-color theme_selected_bg_color @colorAccent;
@define-color textColorPrimary mix(@colorPrimary, #fafafa, 0.95);
@define-color textColorPrimaryShadow alpha(#7239b3, 0.4);

@define-color textColorPrimaryDefault mix(@titlebar_color, @text_color, 0.75);
@define-color textColorPrimaryShadowDefault alpha(shade(@titlebar_color, 1.4), 0.4);

@define-color settings_separator @menu_separator;
@define-color settings_separator_shadow @menu_separator_shadow;

.titlebar.flat .title,
.titlebar.flat .subtitle,
.titlebar.flat .titlebutton,
.titlebar.flat button,
.titlebar.flat label,
.titlebar.flat *:not(entry) image
{
	color: @textColorPrimaryDefault;
	text-shadow: 0 1px @textColorPrimaryShadowDefault;
	-gtk-icon-shadow: 0 1px @textColorPrimaryShadowDefault;
}

.titlebar.flat .title:backdrop,
.titlebar.flat .subtitle:backdrop,
.titlebar.flat .titlebutton:backdrop,
.titlebar.flat button:backdrop,
.titlebar.flat label:backdrop,
.titlebar.flat *:not(entry) image:backdrop
{
	color: mix(@textColorPrimaryDefault, @titlebar_color, 0.3);
	text-shadow: none;
	-gtk-icon-shadow: none;
}

.titlebar:not(.flat) label
{
	color: @textColorPrimary;
	text-shadow: 0 1px @textColorPrimaryShadow;
	-gtk-icon-shadow: 0 1px @textColorPrimaryShadow;
}

.titlebar:not(.flat) label:backdrop
{
	color: mix(@textColorPrimary, @titlebar_color, 0.3);
	text-shadow: none;
	-gtk-icon-shadow: none;
}

infobar.frame.compat-tool-warning
{
	margin: -1px -3px -1px -7px;
}

infobar.frame.settings-info
{
	margin: -1px -3px -1px -3px;
}

.import-emulated-games-dialog .select-all
{
	margin: 0;
	padding: 0;
}
.import-emulated-games-dialog .select-all > check
{
	margin: 0 12px 0 0;
	padding: 0;
}
.import-emulated-games-dialog .select-all > label
{
	margin: 0;
	padding: 0;
}
.import-emulated-games-dialog .dialog-action-box
{
	margin: 0;
	padding: 0 8px 2px 8px;
}
.import-emulated-games-dialog .dialog-action-area
{
	margin: 0;
	padding: 0;
}
